import Logger from '@dzeio/logger'
import { objectMap, objectOmit } from '@dzeio/object-util'
import URLManager from '@dzeio/url-manager'
import type { APIRoute } from 'astro'
import { evaluate } from 'mathjs'
import { exec as execSync, spawn } from 'node:child_process'
import fs from 'node:fs/promises'
import os from 'node:os'
import path from 'node:path'
import { promisify } from 'node:util'
import { buildRFC7807 } from '../../../libs/RFCs/RFC7807'
import { getParams } from '../../../libs/gcodeUtils'
import { validateAuth } from '../../../libs/validateAuth'
import DaoFactory from '../../../models/DaoFactory'

const exec = promisify(execSync)

let tmpDir: string

/**
 * body is the stl
 * query
 * price: algorithm from settings
 * adionnal settings from https://manual.slic3r.org/advanced/command-line
 */
export const post: APIRoute = async ({ params, request }) => {
	const res = await validateAuth(request, {
		name: 'slicing.slice',
		api: true,
		cookie: true
	})
	if (res) {
		return res
	}
	if (!tmpDir) {
		tmpDir = await fs.mkdtemp(path.join(os.tmpdir(), 'saas-'))
	}

	const configId = params.configId ?? 'undefined'
	const config = await DaoFactory.get('config').get(configId)

	if (!config) {
		return buildRFC7807({
			type: '/missing-config',
			status: 404,
			title: 'The configuration does not exists',
			details: `The configuration ${configId} does not exists`
		})
	}

	const query = new URLManager(request.url).query()

	const processId = (Math.random() * 1000000).toFixed(0)
	const logger = new Logger(`process-${processId}`)
	const processFolder = `${tmpDir}/${processId}`
	const pouet = await fs.mkdir(processFolder, { recursive: true })

	logger.log('poeut', pouet)

	logger.log('started processing request')

	logger.log('writing configs to dir')
	for (const file of config.files) {
		await fs.writeFile(`${processFolder}/${file.name}`, file.data)
	}


	const overrides = objectOmit(query, 'algo')

	const stlPath = `${processFolder}/input.stl`
	const gcodePath = `${processFolder}/output.gcode`

	logger.log('writing STL to filesystem')
	// write input
	await fs.writeFile(stlPath, new Uint8Array(Buffer.from(await request.arrayBuffer())), {
		encoding: null
	})

	// additionnal parameters
	let additionnalParams = objectMap(overrides, (value, key) => `--${(key as string).replace(/_/g, '-')} ${value}`).join(' ')


	let slicerPath: string
	let slicerCommand: string

	if (config.type === 'prusa' || true) {
		slicerPath = import.meta.env.PRUSASLICER_PATH ?? 'prusa-slicer'
		additionnalParams += ' --export-gcode'
		slicerCommand = `${path.normalize(stlPath)} --load ${path.normalize(`${processFolder}/config.ini`)} --output ${path.normalize(gcodePath)} ${additionnalParams}`
	}

	try {
		logger.log('Running', slicerPath, slicerCommand)
		await new Promise<void>((res, rej) => {
			const slicer = spawn(slicerPath, slicerCommand.split(' '))
			slicer.stdout.on('data', (data) => {
				logger.log('[stdout]',data.toString('utf8'))
			})
			slicer.stderr.on('data', (data: Buffer) => {
				logger.log('[stderr]', data.toString('utf8'))
			})
			slicer.on('error', (err) => {
				logger.log('error', err)
				rej(err)
			})
			slicer.on('close', (code, signal) => {
				logger.log('code', code)
				logger.log('signal', signal)
				if (typeof code === 'number' && code > 0) {

					rej(code)
					return
				}
				res()
			})
		})
	} catch (e: any) {
		logger.log('request finished in error :(', processId)
		const line = e.toString()
		logger.error(e)
		if (line.includes('Objects could not fit on the bed')) {
			await fs.rm(stlPath)
			return buildRFC7807({
				type: '/object-too-large',
				status: 413,
				title: 'Object is too large'
			})
		} else if (line.includes('No such file')) {
			await fs.rm(stlPath)
			return buildRFC7807({
				type: '/missing-config-file',
				status: 404,
				title: 'Configuration file is missing',
				details: `the configuration file "${configId}" is not available on the remote server`
			})
		} else if (line.includes('Unknown option')) {
			await fs.rm(stlPath)
			return buildRFC7807({
				type: '/slicer-option-unknown',
				status: 400,
				title: ' config override doew not exists',
				details: 'an override does not exists, please contact an administrator or refer to the documentation'
			})
		} else if (
			line.includes('is not recognized as an internal or external command') ||
			line.includes('.dll was not loaded')
		) {
			await fs.rm(stlPath)
			return buildRFC7807({
				type: '/slicer-not-found',
				status: 408,
				title: 'the slicer used to process this file has not been found',
				details: 'the server has a misconfiguration meaning that we can\'t process the request, please contact an administrator',
				additionnalInfo: line.includes('dll') ? 'Missing DLL' : 'Slicer not found '
			})
		} else if (line.includes('ETIMEDOUT')) {
			await fs.rm(stlPath)
			return buildRFC7807({
				type: '/timed-out-slicing',
				status: 408,
				title: 'Timed out slicing file',
				detail: `The file you are trying to process takes too long to be processed`,
				processingTimeoutMillis: 60000
			})
		}
		return buildRFC7807({
			type: '/general-input-output-error',
			status: 500,
			title: 'General I/O error',
			details: 'A server error make it impossible to slice the file, please contact an administrator with the json error',
			fileId: processId,
			config: configId,
			// fileSize: req.body.length,
			overrides: overrides,
			serverMessage:
				e.toString().replace(new RegExp(stlPath), `***FILE***`).replace(new RegExp(processFolder), '')
		})
	}
	const gcode = await fs.readFile(gcodePath, 'utf-8')
	await fs.rm(processFolder, { recursive: true, force: true })
	logger.log('Getting parameters')
	const gcodeParams = getParams(gcode)
	let price: string | undefined
	if (query?.algo) {
		let algo = decodeURI(query.algo as string)
		// objectLoop(params, (value, key) => {
		// 	if (typeof value !== 'number') {
		// 		return
		// 	}
		// 	while (algo.includes(key)) {
		// 		algo = algo.replace(key, value.toString())
		// 	}
		// })
		try {
			logger.log('Evaluating Alogrithm')
			const tmp = evaluate(algo, gcodeParams)
			if (typeof tmp === 'number') {
				price = tmp.toFixed(2)
			} else {
				return buildRFC7807({
					type: '/algorithm-error',
					status: 500,
					title: 'Algorithm compilation error',
					details: 'It seems the algorithm resolution failed',
					algorithm: algo,
					algorithmError: 'Algorithm return a Unit',
					parameters: gcodeParams
				})
			}
		} catch (e) {
			logger.dir(e)
			return buildRFC7807({
				type: '/algorithm-error',
				status: 500,
				title: 'Algorithm compilation error',
				details: 'It seems the algorithm resolution failed',
				algorithm: algo,
				algorithmError: e,
				parameters: gcodeParams
			})
		}
	}
	logger.log('request successfull :)')
	return {
		status: 200,
		body: JSON.stringify({
			price: price ? parseFloat(price) : undefined,
			...getParams(gcode),
			gcode
		})
	}
}
